# Generated by Plain 0.21.5 on 2025-02-17 04:24

from plain import models
from plain.models import migrations


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="User",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True)),
                ("email", models.EmailField(max_length=254)),
                ("username", models.CharField(max_length=100)),
            ],
        ),
    ]



from os import environ

SECRET_KEY = "test"
URLS_ROUTER = "app.urls.AppRouter"
INSTALLED_PACKAGES = [
    "plain.auth",
    "plain.sessions",
    "plain.models",
    "plain.oauth",
    "app.users",
]
DATABASES = {
    "default": {
        "ENGINE": "plain.models.backends.sqlite3",
        "NAME": ":memory:",
    }
}
MIDDLEWARE = [
    "plain.sessions.middleware.SessionMiddleware",
    "plain.auth.middleware.AuthenticationMiddleware",
]
AUTH_LOGIN_URL = "login"
AUTH_USER_MODEL = "users.User"

# OAuth providers to use for a real, interactive test
# (in a real config you'd probably do environ["key"] to raise a KeyError if an env var is forgotten)
OAUTH_LOGIN_PROVIDERS = {
    "github": {
        "class": "providers.github.GitHubOAuthProvider",
        "kwargs": {
            "client_id": environ.get("GITHUB_CLIENT_ID"),
            "client_secret": environ.get("GITHUB_CLIENT_SECRET"),
        },
    },
    "bitbucket": {
        "class": "providers.bitbucket.BitbucketOAuthProvider",
        "kwargs": {
            "client_id": environ.get("BITBUCKET_KEY"),
            "client_secret": environ.get("BITBUCKET_SECRET"),
        },
    },
    "gitlab": {
        "class": "providers.gitlab.GitLabOAuthProvider",
        "kwargs": {
            "client_id": environ.get("GITLAB_APPLICATION_ID"),
            "client_secret": environ.get("GITLAB_APPLICATION_SECRET"),
            "scope": "read_user",
        },
    },
}



# Generated by Plain 0.3.0 on 2024-08-16 16:46

from plain import models
from plain.models import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("users", "__first__"),
        ("plainoauth", "0004_alter_oauthconnection_access_token_and_more"),
    ]

    operations = [
        migrations.AddConstraint(
            model_name="oauthconnection",
            constraint=models.UniqueConstraint(
                fields=("provider_key", "provider_user_id"),
                name="unique_oauth_provider_user_id",
            ),
        ),
    ]



import requests

from plain.oauth.providers import OAuthProvider, OAuthToken, OAuthUser


class GitLabOAuthProvider(OAuthProvider):
    authorization_url = "https://gitlab.com/oauth/authorize"

    def _get_token(self, request_data):
        request_data["client_id"] = self.get_client_id()
        request_data["client_secret"] = self.get_client_secret()
        response = requests.post(
            "https://gitlab.com/oauth/token",
            headers={
                "Accept": "application/json",
            },
            data=request_data,
        )
        response.raise_for_status()
        data = response.json()
        return OAuthToken(
            access_token=data["access_token"],
            refresh_token=data["refresh_token"],
            # expires_in is missing in response?
        )

    def get_oauth_token(self, *, code, request):
        return self._get_token(
            {
                "grant_type": "authorization_code",
                "code": code,
                "redirect_uri": self.get_callback_url(request=request),
            }
        )

    def refresh_oauth_token(self, *, oauth_token):
        return self._get_token(
            {
                "grant_type": "refresh_token",
                "refresh_token": oauth_token.refresh_token,
            }
        )

    def get_oauth_user(self, *, oauth_token):
        response = requests.get(
            "https://gitlab.com/api/v4/user",
            headers={
                "Authorization": f"Bearer {oauth_token.access_token}",
            },
        )
        response.raise_for_status()
        data = response.json()
        return OAuthUser(
            id=data["id"],
            email=data["email"],
            username=data["username"],
        )



# Generated by Plain 0.31.0 on 2025-03-08 21:33

from plain import models
from plain.models import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("plainoauth", "0005_alter_oauthconnection_unique_together_and_more"),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name="oauthconnection",
            name="unique_oauth_provider_user_id",
        ),
        migrations.AddConstraint(
            model_name="oauthconnection",
            constraint=models.UniqueConstraint(
                fields=("provider_key", "provider_user_id"),
                name="plainoauth_oauthconnection_unique_provider_key_user_id",
            ),
        ),
    ]



import datetime

import requests

from plain.oauth.exceptions import OAuthError
from plain.oauth.providers import OAuthProvider, OAuthToken, OAuthUser
from plain.utils import timezone


class GitHubOAuthProvider(OAuthProvider):
    authorization_url = "https://github.com/login/oauth/authorize"

    github_token_url = "https://github.com/login/oauth/access_token"
    github_user_url = "https://api.github.com/user"
    github_emails_url = "https://api.github.com/user/emails"

    def _get_token(self, request_data):
        response = requests.post(
            self.github_token_url,
            headers={
                "Accept": "application/json",
            },
            data=request_data,
        )
        response.raise_for_status()
        data = response.json()

        oauth_token = OAuthToken(
            access_token=data["access_token"],
        )

        # Expiration and refresh tokens are optional in GitHub depending on the app
        if "expires_in" in data:
            oauth_token.access_token_expires_at = timezone.now() + datetime.timedelta(
                seconds=data["expires_in"]
            )

        if "refresh_token" in data:
            oauth_token.refresh_token = data["refresh_token"]

        if "refresh_token_expires_in" in data:
            oauth_token.refresh_token_expires_at = timezone.now() + datetime.timedelta(
                seconds=data["refresh_token_expires_in"]
            )

        return oauth_token

    def get_oauth_token(self, *, code, request):
        return self._get_token(
            {
                "client_id": self.get_client_id(),
                "client_secret": self.get_client_secret(),
                "code": code,
            }
        )

    def refresh_oauth_token(self, *, oauth_token):
        return self._get_token(
            {
                "client_id": self.get_client_id(),
                "client_secret": self.get_client_secret(),
                "refresh_token": oauth_token.refresh_token,
                "grant_type": "refresh_token",
            }
        )

    def get_oauth_user(self, *, oauth_token):
        response = requests.get(
            self.github_user_url,
            headers={
                "Accept": "application/json",
                "Authorization": f"token {oauth_token.access_token}",
            },
        )
        response.raise_for_status()
        data = response.json()
        user_id = data["id"]
        username = data["login"]

        # Use the verified, primary email address (not the public profile email, which is optional anyway)
        response = requests.get(
            self.github_emails_url,
            headers={
                "Accept": "application/json",
                "Authorization": f"token {oauth_token.access_token}",
            },
        )
        response.raise_for_status()

        try:
            verified_primary_email = [
                x["email"] for x in response.json() if x["primary"] and x["verified"]
            ][0]
        except IndexError:
            raise OAuthError("A verified primary email address is required on GitHub")

        return OAuthUser(
            id=user_id,
            email=verified_primary_email,
            username=username,
        )



import datetime

import requests

from plain.oauth.providers import OAuthProvider, OAuthToken, OAuthUser
from plain.utils import timezone


class BitbucketOAuthProvider(OAuthProvider):
    authorization_url = "https://bitbucket.org/site/oauth2/authorize"

    def _get_token(self, request_data):
        response = requests.post(
            "https://bitbucket.org/site/oauth2/access_token",
            auth=(self.get_client_id(), self.get_client_secret()),
            headers={
                "Accept": "application/json",
            },
            data=request_data,
        )
        response.raise_for_status()
        data = response.json()
        return OAuthToken(
            access_token=data["access_token"],
            refresh_token=data["refresh_token"],
            access_token_expires_at=timezone.now()
            + datetime.timedelta(seconds=data["expires_in"]),
        )

    def get_oauth_token(self, *, code, request):
        return self._get_token(
            {
                "grant_type": "authorization_code",
                "code": code,
                "redirect_uri": self.get_callback_url(request=request),
            }
        )

    def refresh_oauth_token(self, *, oauth_token):
        return self._get_token(
            {
                "grant_type": "refresh_token",
                "refresh_token": oauth_token.refresh_token,
            }
        )

    def get_oauth_user(self, *, oauth_token):
        response = requests.get(
            "https://api.bitbucket.org/2.0/user",
            headers={
                "Authorization": f"Bearer {oauth_token.access_token}",
            },
        )
        response.raise_for_status()
        user_id = response.json()["uuid"]
        username = response.json()["username"]

        response = requests.get(
            "https://api.bitbucket.org/2.0/user/emails",
            headers={
                "Authorization": f"Bearer {oauth_token.access_token}",
            },
        )
        response.raise_for_status()
        confirmed_primary_email = [
            x["email"]
            for x in response.json()["values"]
            if x["is_primary"] and x["is_confirmed"]
        ][0]

        return OAuthUser(
            id=user_id,
            email=confirmed_primary_email,
            username=username,
        )



from plain import models


@models.register_model
class User(models.Model):
    email = models.EmailField()
    username = models.CharField(max_length=100)

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=["email"], name="user_unique_email"),
            models.UniqueConstraint(fields=["username"], name="user_unique_username"),
        ]

    def __str__(self):
        return self.username



# Generated by Plain 0.32.0 on 2025-03-10 15:37

from plain import models
from plain.models import migrations


class Migration(migrations.Migration):
    dependencies = [
        (
            "plainoauth",
            "0006_remove_oauthconnection_unique_oauth_provider_user_id_and_more",
        ),
    ]

    operations = [
        migrations.AlterField(
            model_name="oauthconnection",
            name="provider_key",
            field=models.CharField(max_length=100),
        ),
        migrations.AlterField(
            model_name="oauthconnection",
            name="provider_user_id",
            field=models.CharField(max_length=100),
        ),
    ]



from plain.auth import get_user_model
from plain.oauth.models import OAuthConnection


def test_oauth_provider_keys_check_pass(db, settings):
    settings.OAUTH_LOGIN_PROVIDERS = {
        "google": {
            "client_id": "test_client_id",
            "client_secret": "test_client_secret",
        },
        "foo": {
            "client_id": "test_client_id",
            "client_secret": "test_client_secret",
        },
    }

    user = get_user_model().objects.create(
        username="test_user", email="test@example.com"
    )

    OAuthConnection.objects.create(
        user=user,
        provider_key="google",
        provider_user_id="test_provider_user_id",
        access_token="test",
    )

    errors = OAuthConnection.check(databases=["default"])
    assert len(errors) == 0


def test_oauth_provider_keys_check_fail(db, settings):
    settings.OAUTH_LOGIN_PROVIDERS = {
        "google": {
            "client_id": "test_client_id",
            "client_secret": "test_client_secret",
        },
        "foo": {
            "client_id": "test_client_id",
            "client_secret": "test_client_secret",
        },
    }

    user = get_user_model().objects.create(
        username="test_user", email="test@example.com"
    )

    OAuthConnection.objects.create(
        user=user,
        provider_key="google",
        provider_user_id="test_provider_user_id",
        access_token="test",
    )
    OAuthConnection.objects.create(
        user=user,
        provider_key="bar",
        provider_user_id="test_provider_user_id",
        access_token="test",
    )

    errors = OAuthConnection.check(databases=["default"])
    assert len(errors) == 1
    assert (
        errors[0].msg
        == "The following OAuth providers are in the database but not in the settings: bar"
    )



import datetime

from plain.auth import get_user_model
from plain.oauth.models import OAuthConnection
from plain.oauth.providers import OAuthProvider, OAuthToken, OAuthUser
from plain.test import Client


class DummyProvider(OAuthProvider):
    authorization_url = "https://example.com/oauth/authorize"

    def generate_state(self) -> str:
        return "dummy_state"

    def refresh_oauth_token(self, *, oauth_token: OAuthToken) -> OAuthToken:
        return OAuthToken(
            access_token="refreshed_dummy_access_token",
            refresh_token="refreshed_dummy_refresh_token",
            access_token_expires_at=datetime.datetime(
                2029, 1, 1, 0, 0, tzinfo=datetime.UTC
            ),
            refresh_token_expires_at=datetime.datetime(
                2029, 1, 2, 0, 0, tzinfo=datetime.UTC
            ),
        )

    def get_oauth_token(self, *, code, request) -> OAuthToken:
        return OAuthToken(
            access_token="dummy_access_token",
            refresh_token="dummy_refresh_token",
            access_token_expires_at=datetime.datetime(
                2020, 1, 1, 0, 0, tzinfo=datetime.UTC
            ),
            refresh_token_expires_at=datetime.datetime(
                2020, 1, 2, 0, 0, tzinfo=datetime.UTC
            ),
        )

    def get_oauth_user(self, *, oauth_token: OAuthToken) -> OAuthUser:
        return OAuthUser(
            id="dummy_id",
            email="dummy@example.com",
            username="dummy_username",
        )


def test_dummy_signup(db, settings):
    settings.OAUTH_LOGIN_PROVIDERS = {
        "dummy": {
            "class": "test_providers.DummyProvider",
            "kwargs": {
                "client_id": "dummy_client_id",
                "client_secret": "dummy_client_secret",
                "scope": "dummy_scope",
            },
        }
    }

    client = Client()

    assert get_user_model().objects.count() == 0
    assert OAuthConnection.objects.count() == 0

    # Login required for this view
    response = client.get("/")
    assert response.status_code == 302
    assert response.url == "/login/?next=/"

    # User clicks the login link (form submit)
    response = client.post("/oauth/dummy/login/")
    assert response.status_code == 302
    assert (
        response.url
        == "https://example.com/oauth/authorize?client_id=dummy_client_id&redirect_uri=https%3A%2F%2Ftestserver%2Foauth%2Fdummy%2Fcallback%2F&response_type=code&scope=dummy_scope&state=dummy_state"
    )

    # Provider redirects to the callback url
    response = client.get("/oauth/dummy/callback/?code=test_code&state=dummy_state")
    assert response.status_code == 302
    assert response.url == "/"

    # Now logged in
    response = client.get("/")
    assert response.status_code == 200
    assert b"Hello dummy_username!\n" in response.content

    # Check the user and connection that was created
    user = response.user
    assert user.username == "dummy_username"
    assert user.email == "dummy@example.com"
    connections = user.oauth_connections.all()
    assert len(connections) == 1
    assert connections[0].provider_key == "dummy"
    assert connections[0].provider_user_id == "dummy_id"
    assert connections[0].access_token == "dummy_access_token"
    assert connections[0].refresh_token == "dummy_refresh_token"
    assert connections[0].access_token_expires_at == datetime.datetime(
        2020, 1, 1, 0, 0, tzinfo=datetime.UTC
    )
    assert connections[0].refresh_token_expires_at == datetime.datetime(
        2020, 1, 2, 0, 0, tzinfo=datetime.UTC
    )

    assert get_user_model().objects.count() == 1
    assert OAuthConnection.objects.count() == 1


def test_dummy_login_connection(db, settings):
    settings.OAUTH_LOGIN_PROVIDERS = {
        "dummy": {
            "class": "test_providers.DummyProvider",
            "kwargs": {
                "client_id": "dummy_client_id",
                "client_secret": "dummy_client_secret",
                "scope": "dummy_scope",
            },
        }
    }

    client = Client()

    assert get_user_model().objects.count() == 0
    assert OAuthConnection.objects.count() == 0

    # Create a user
    user = get_user_model().objects.create(
        username="dummy_username", email="dummy@example.com"
    )
    OAuthConnection.objects.create(
        user=user,
        provider_key="dummy",
        provider_user_id="dummy_id",
        access_token="dummy_access_token",
        refresh_token="dummy_refresh_token",
        access_token_expires_at=datetime.datetime(
            2020, 1, 1, 0, 0, tzinfo=datetime.UTC
        ),
        refresh_token_expires_at=datetime.datetime(
            2020, 1, 2, 0, 0, tzinfo=datetime.UTC
        ),
    )

    assert get_user_model().objects.count() == 1
    assert OAuthConnection.objects.count() == 1

    # Login required for this view
    response = client.get("/")
    assert response.status_code == 302
    assert response.url == "/login/?next=/"

    # User clicks the login link (form submit)
    response = client.post("/oauth/dummy/login/")
    assert response.status_code == 302
    assert (
        response.url
        == "https://example.com/oauth/authorize?client_id=dummy_client_id&redirect_uri=https%3A%2F%2Ftestserver%2Foauth%2Fdummy%2Fcallback%2F&response_type=code&scope=dummy_scope&state=dummy_state"
    )

    # Provider redirects to the callback url
    response = client.get("/oauth/dummy/callback/?code=test_code&state=dummy_state")
    assert response.status_code == 302
    assert response.url == "/"

    # Now logged in
    response = client.get("/")
    assert response.status_code == 200
    assert b"Hello dummy_username!\n" in response.content

    # Check the user and connection that was created
    user = response.user
    assert user.username == "dummy_username"
    assert user.email == "dummy@example.com"
    connections = user.oauth_connections.all()
    assert len(connections) == 1
    assert connections[0].provider_key == "dummy"
    assert connections[0].provider_user_id == "dummy_id"
    assert connections[0].access_token == "dummy_access_token"
    assert connections[0].refresh_token == "dummy_refresh_token"
    assert connections[0].access_token_expires_at == datetime.datetime(
        2020, 1, 1, 0, 0, tzinfo=datetime.UTC
    )
    assert connections[0].refresh_token_expires_at == datetime.datetime(
        2020, 1, 2, 0, 0, tzinfo=datetime.UTC
    )

    assert get_user_model().objects.count() == 1
    assert OAuthConnection.objects.count() == 1


def test_dummy_login_without_connection(db, settings):
    settings.OAUTH_LOGIN_PROVIDERS = {
        "dummy": {
            "class": "test_providers.DummyProvider",
            "kwargs": {
                "client_id": "dummy_client_id",
                "client_secret": "dummy_client_secret",
                "scope": "dummy_scope",
            },
        }
    }

    client = Client()

    assert get_user_model().objects.count() == 0
    assert OAuthConnection.objects.count() == 0

    # Create a user
    get_user_model().objects.create(
        username="dummy_username", email="dummy@example.com"
    )

    assert get_user_model().objects.count() == 1
    assert OAuthConnection.objects.count() == 0

    # Login required for this view
    response = client.get("/")
    assert response.status_code == 302
    assert response.url == "/login/?next=/"

    # User clicks the login link (form submit)
    response = client.post("/oauth/dummy/login/")
    assert response.status_code == 302
    assert (
        response.url
        == "https://example.com/oauth/authorize?client_id=dummy_client_id&redirect_uri=https%3A%2F%2Ftestserver%2Foauth%2Fdummy%2Fcallback%2F&response_type=code&scope=dummy_scope&state=dummy_state"
    )

    # Provider redirects to the callback url
    response = client.get("/oauth/dummy/callback/?code=test_code&state=dummy_state")
    assert response.status_code == 400
    assert b"OAuth Error" in response.content


def test_dummy_connect(db, settings):
    settings.OAUTH_LOGIN_PROVIDERS = {
        "dummy": {
            "class": "test_providers.DummyProvider",
            "kwargs": {
                "client_id": "dummy_client_id",
                "client_secret": "dummy_client_secret",
                "scope": "dummy_scope",
            },
        }
    }

    client = Client()

    assert get_user_model().objects.count() == 0
    assert OAuthConnection.objects.count() == 0

    # Create a user
    user = get_user_model().objects.create(
        username="dummy_username", email="dummy@example.com"
    )

    assert get_user_model().objects.count() == 1
    assert OAuthConnection.objects.count() == 0

    client.force_login(user)

    response = client.post("/oauth/dummy/connect/")
    assert response.status_code == 302
    assert (
        response.url
        == "https://example.com/oauth/authorize?client_id=dummy_client_id&redirect_uri=https%3A%2F%2Ftestserver%2Foauth%2Fdummy%2Fcallback%2F&response_type=code&scope=dummy_scope&state=dummy_state"
    )

    # Provider redirects to the callback url
    response = client.get("/oauth/dummy/callback/?code=test_code&state=dummy_state")
    assert response.status_code == 302
    assert response.url == "/"

    # Now logged in
    response = client.get("/")

    # Check the user and connection that was created
    user = response.user
    connections = user.oauth_connections.all()
    assert len(connections) == 1
    assert connections[0].provider_key == "dummy"
    assert connections[0].provider_user_id == "dummy_id"
    assert connections[0].access_token == "dummy_access_token"
    assert connections[0].refresh_token == "dummy_refresh_token"
    assert connections[0].access_token_expires_at == datetime.datetime(
        2020, 1, 1, 0, 0, tzinfo=datetime.UTC
    )
    assert connections[0].refresh_token_expires_at == datetime.datetime(
        2020, 1, 2, 0, 0, tzinfo=datetime.UTC
    )

    assert get_user_model().objects.count() == 1
    assert OAuthConnection.objects.count() == 1


# def test_dummy_disconnect_to_password(db, client, settings):
#     settings.OAUTH_LOGIN_PROVIDERS = {
#         "dummy": {
#             "class": "test_providers.DummyProvider",
#             "kwargs": {
#                 "client_id": "dummy_client_id",
#                 "client_secret": "dummy_client_secret",
#                 "scope": "dummy_scope",
#             },
#         }
#     }

#     assert get_user_model().objects.count() == 0
#     assert OAuthConnection.objects.count() == 0

#     # Create a user
#     user = get_user_model().objects.create(
#         username="dummy_username", email="dummy@example.com", password="dummy_password"
#     )
#     OAuthConnection.objects.create(
#         user=user,
#         provider_key="dummy",
#         provider_user_id="dummy_id",
#         access_token="dummy_access_token",
#         refresh_token="dummy_refresh_token",
#         access_token_expires_at=datetime.datetime(
#             2020, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
#         ),
#         refresh_token_expires_at=datetime.datetime(
#             2020, 1, 2, 0, 0, tzinfo=datetime.timezone.utc
#         ),
#     )

#     assert get_user_model().objects.count() == 1
#     assert OAuthConnection.objects.count() == 1

#     client.force_login(user)

#     # Raises a BadRequest error - can't disconnect the last connection without a password
#     response = client.post(
#         "/oauth/dummy/disconnect/", data={"provider_user_id": "dummy_id"}
#     )
#     assert response.status_code == 302
#     assert response.url == "/"

#     assert get_user_model().objects.count() == 1
#     assert OAuthConnection.objects.count() == 0


# def test_dummy_disconnect_to_connection(db, client, settings):
#     settings.OAUTH_LOGIN_PROVIDERS = {
#         "dummy": {
#             "class": "test_providers.DummyProvider",
#             "kwargs": {
#                 "client_id": "dummy_client_id",
#                 "client_secret": "dummy_client_secret",
#                 "scope": "dummy_scope",
#             },
#         }
#     }

#     assert get_user_model().objects.count() == 0
#     assert OAuthConnection.objects.count() == 0

#     # Create a user
#     user = get_user_model().objects.create(
#         username="dummy_username", email="dummy@example.com"
#     )
#     OAuthConnection.objects.create(
#         user=user,
#         provider_key="dummy",
#         provider_user_id="dummy_id",
#         access_token="dummy_access_token",
#         refresh_token="dummy_refresh_token",
#         access_token_expires_at=datetime.datetime(
#             2020, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
#         ),
#         refresh_token_expires_at=datetime.datetime(
#             2020, 1, 2, 0, 0, tzinfo=datetime.timezone.utc
#         ),
#     )
#     OAuthConnection.objects.create(
#         user=user,
#         provider_key="dummy",
#         provider_user_id="dummy_id2",
#         access_token="dummy_access_token",
#         refresh_token="dummy_refresh_token",
#         access_token_expires_at=datetime.datetime(
#             2020, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
#         ),
#         refresh_token_expires_at=datetime.datetime(
#             2020, 1, 2, 0, 0, tzinfo=datetime.timezone.utc
#         ),
#     )

#     assert get_user_model().objects.count() == 1
#     assert OAuthConnection.objects.count() == 2

#     client.force_login(user)

#     # Raises a BadRequest error - can't disconnect the last connection without a password
#     response = client.post(
#         "/oauth/dummy/disconnect/", data={"provider_user_id": "dummy_id"}
#     )
#     assert response.status_code == 302
#     assert response.url == "/"

#     assert get_user_model().objects.count() == 1
#     assert OAuthConnection.objects.count() == 1


# def test_dummy_disconnect_last(db, client, settings):
#     settings.OAUTH_LOGIN_PROVIDERS = {
#         "dummy": {
#             "class": "test_providers.DummyProvider",
#             "kwargs": {
#                 "client_id": "dummy_client_id",
#                 "client_secret": "dummy_client_secret",
#                 "scope": "dummy_scope",
#             },
#         }
#     }

#     assert get_user_model().objects.count() == 0
#     assert OAuthConnection.objects.count() == 0

#     # Create a user
#     user = get_user_model().objects.create(
#         username="dummy_username", email="dummy@example.com"
#     )
#     OAuthConnection.objects.create(
#         user=user,
#         provider_key="dummy",
#         provider_user_id="dummy_id",
#         access_token="dummy_access_token",
#         refresh_token="dummy_refresh_token",
#         access_token_expires_at=datetime.datetime(
#             2020, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
#         ),
#         refresh_token_expires_at=datetime.datetime(
#             2020, 1, 2, 0, 0, tzinfo=datetime.timezone.utc
#         ),
#     )

#     assert get_user_model().objects.count() == 1
#     assert OAuthConnection.objects.count() == 1

#     client.force_login(user)

#     # Raises a BadRequest error - can't disconnect the last connection without a password
#     response = client.post(
#         "/oauth/dummy/disconnect/", data={"provider_user_id": "dummy_id"}
#     )
#     assert response.status_code == 400
#     assert response.templates[0].name == "oauth/error.html"

#     assert get_user_model().objects.count() == 1
#     assert OAuthConnection.objects.count() == 1


def test_dummy_refresh(db, settings, monkeypatch):
    settings.OAUTH_LOGIN_PROVIDERS = {
        "dummy": {
            "class": "test_providers.DummyProvider",
            "kwargs": {
                "client_id": "dummy_client_id",
                "client_secret": "dummy_client_secret",
                "scope": "dummy_scope",
            },
        }
    }

    user = get_user_model().objects.create(
        username="dummy_username", email="dummy@example.com"
    )
    connection = OAuthConnection.objects.create(
        user=user,
        provider_key="dummy",
        provider_user_id="dummy_id",
        access_token="dummy_access_token",
        refresh_token="dummy_refresh_token",
        access_token_expires_at=datetime.datetime(
            2020, 1, 1, 0, 0, tzinfo=datetime.UTC
        ),
        refresh_token_expires_at=datetime.datetime(
            2020, 1, 2, 0, 0, tzinfo=datetime.UTC
        ),
    )

    connection.refresh_access_token()
    assert connection.provider_key == "dummy"
    assert connection.provider_user_id == "dummy_id"
    assert connection.access_token == "refreshed_dummy_access_token"
    assert connection.refresh_token == "refreshed_dummy_refresh_token"
    assert connection.access_token_expires_at == datetime.datetime(
        2029, 1, 1, 0, 0, tzinfo=datetime.UTC
    )
    assert connection.refresh_token_expires_at == datetime.datetime(
        2029, 1, 2, 0, 0, tzinfo=datetime.UTC
    )



from plain.oauth.providers import OAuthProvider, OAuthToken, OAuthUser
from plain.test import Client


class DummyProvider(OAuthProvider):
    def get_oauth_token(self, *, code, request):
        return OAuthToken(
            access_token="dummy_token",
        )

    def get_oauth_user(self, *, oauth_token):
        return OAuthUser(
            id="dummy_user_id",
            username="dummy_username",
            email="dummy@example.com",
        )

    def check_request_state(self, *, request):
        """Don't check the state"""
        return


def test_single_backend(db, settings):
    settings.OAUTH_LOGIN_PROVIDERS = {
        "dummy": {
            "class": "test_backends.DummyProvider",
            "kwargs": {
                "client_id": "dummy_client_id",
                "client_secret": "dummy_client_secret",
                "scope": "dummy_scope",
            },
        }
    }

    client = Client()

    response = client.get("/oauth/dummy/callback/?code=test_code&state=dummy_state")
    assert response.status_code == 302
    assert response.url == "/"

    # Now logged in
    response = client.get("/")
    assert response.user


def test_multiple_backends(db, settings):
    settings.OAUTH_LOGIN_PROVIDERS = {
        "dummy": {
            "class": "test_backends.DummyProvider",
            "kwargs": {
                "client_id": "dummy_client_id",
                "client_secret": "dummy_client_secret",
                "scope": "dummy_scope",
            },
        }
    }

    client = Client()

    response = client.get("/oauth/dummy/callback/?code=test_code&state=dummy_state")
    assert response.status_code == 302
    assert response.url == "/"

    # Now logged in
    response = client.get("/")
    assert response.user



from plain.auth.views import AuthViewMixin, LogoutView
from plain.oauth.providers import get_provider_keys
from plain.oauth.urls import OAuthRouter
from plain.urls import Router, include, path
from plain.views import TemplateView


class LoggedInView(AuthViewMixin, TemplateView):
    template_name = "index.html"

    def get_template_context(self):
        context = super().get_template_context()
        context["oauth_provider_keys"] = get_provider_keys()
        return context


class LoginView(TemplateView):
    template_name = "login.html"


class AppRouter(Router):
    namespace = ""
    urls = [
        include("oauth/", OAuthRouter),
        path("login/", LoginView, name="login"),
        path("logout/", LogoutView, name="logout"),
        path("", LoggedInView),
    ]



from typing import TYPE_CHECKING

from plain import models
from plain.auth import get_user_model
from plain.exceptions import ValidationError
from plain.models import transaction
from plain.models.db import IntegrityError, OperationalError, ProgrammingError
from plain.preflight import Error
from plain.runtime import SettingsReference
from plain.utils import timezone

from .exceptions import OAuthUserAlreadyExistsError

if TYPE_CHECKING:
    from .providers import OAuthToken, OAuthUser


# TODO preflight check for deploy that ensures all provider keys in db are also in settings?


@models.register_model
class OAuthConnection(models.Model):
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    user = models.ForeignKey(
        SettingsReference("AUTH_USER_MODEL"),
        on_delete=models.CASCADE,
        related_name="oauth_connections",
    )

    # The key used to refer to this provider type (in settings)
    provider_key = models.CharField(max_length=100)

    # The unique ID of the user on the provider's system
    provider_user_id = models.CharField(max_length=100)

    # Token data
    access_token = models.CharField(max_length=2000)
    refresh_token = models.CharField(max_length=2000, required=False)
    access_token_expires_at = models.DateTimeField(required=False, allow_null=True)
    refresh_token_expires_at = models.DateTimeField(required=False, allow_null=True)

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["provider_key", "provider_user_id"],
                name="plainoauth_oauthconnection_unique_provider_key_user_id",
            )
        ]
        ordering = ("provider_key",)

    def __str__(self):
        return f"{self.provider_key}[{self.user}:{self.provider_user_id}]"

    def refresh_access_token(self) -> None:
        from .providers import OAuthToken, get_oauth_provider_instance

        provider_instance = get_oauth_provider_instance(provider_key=self.provider_key)
        oauth_token = OAuthToken(
            access_token=self.access_token,
            refresh_token=self.refresh_token,
            access_token_expires_at=self.access_token_expires_at,
            refresh_token_expires_at=self.refresh_token_expires_at,
        )
        refreshed_oauth_token = provider_instance.refresh_oauth_token(
            oauth_token=oauth_token
        )
        self.set_token_fields(refreshed_oauth_token)
        self.save()

    def set_token_fields(self, oauth_token: "OAuthToken"):
        self.access_token = oauth_token.access_token
        self.refresh_token = oauth_token.refresh_token
        self.access_token_expires_at = oauth_token.access_token_expires_at
        self.refresh_token_expires_at = oauth_token.refresh_token_expires_at

    def set_user_fields(self, oauth_user: "OAuthUser"):
        self.provider_user_id = oauth_user.id

    def access_token_expired(self) -> bool:
        return (
            self.access_token_expires_at is not None
            and self.access_token_expires_at < timezone.now()
        )

    def refresh_token_expired(self) -> bool:
        return (
            self.refresh_token_expires_at is not None
            and self.refresh_token_expires_at < timezone.now()
        )

    @classmethod
    def get_or_create_user(
        cls, *, provider_key: str, oauth_token: "OAuthToken", oauth_user: "OAuthUser"
    ) -> "OAuthConnection":
        try:
            connection = cls.objects.get(
                provider_key=provider_key,
                provider_user_id=oauth_user.id,
            )
            connection.set_token_fields(oauth_token)
            connection.save()
            return connection
        except cls.DoesNotExist:
            with transaction.atomic():
                # If email needs to be unique, then we expect
                # that to be taken care of on the user model itself
                try:
                    user = get_user_model()(
                        **oauth_user.user_model_fields,
                    )
                    user.save()
                except (IntegrityError, ValidationError):
                    raise OAuthUserAlreadyExistsError()

                return cls.connect(
                    user=user,
                    provider_key=provider_key,
                    oauth_token=oauth_token,
                    oauth_user=oauth_user,
                )

    @classmethod
    def connect(
        cls,
        *,
        user,
        provider_key: str,
        oauth_token: "OAuthToken",
        oauth_user: "OAuthUser",
    ) -> "OAuthConnection":
        """
        Connect will either create a new connection or update an existing connection
        """
        try:
            connection = cls.objects.get(
                user=user,
                provider_key=provider_key,
                provider_user_id=oauth_user.id,
            )
        except cls.DoesNotExist:
            # Create our own instance (not using get_or_create)
            # so that any created signals contain the token fields too
            connection = cls(
                user=user,
                provider_key=provider_key,
                provider_user_id=oauth_user.id,
            )

        connection.set_user_fields(oauth_user)
        connection.set_token_fields(oauth_token)
        connection.save()

        return connection

    @classmethod
    def check(cls, **kwargs):
        """
        A system check for ensuring that provider_keys in the database are also present in settings.

        Note that the --database flag is required for this to work:
          python manage.py check --database default
        """
        errors = super().check(**kwargs)

        databases = kwargs.get("databases", None)
        if not databases:
            return errors

        from .providers import get_provider_keys

        for database in databases:
            try:
                keys_in_db = set(
                    cls.objects.using(database)
                    .values_list("provider_key", flat=True)
                    .distinct()
                )
            except (OperationalError, ProgrammingError):
                # Check runs on manage.py migrate, and the table may not exist yet
                # or it may not be installed on the particular database intentionally
                continue

            keys_in_settings = set(get_provider_keys())

            if keys_in_db - keys_in_settings:
                errors.append(
                    Error(
                        "The following OAuth providers are in the database but not in the settings: {}".format(
                            ", ".join(keys_in_db - keys_in_settings)
                        ),
                        id="plain.oauth.E001",
                    )
                )

        return errors



from tests.providers.github import GitHubOAuthProvider

from plain.oauth.providers import OAuthToken, OAuthUser
from plain.test import Client


class DummyGitHubOAuthProvider(GitHubOAuthProvider):
    def generate_state(self) -> str:
        return "dummy_state"

    def get_oauth_token(self, code, request):
        return OAuthToken(access_token="gho_key")

    def get_oauth_user(self, oauth_token):
        return OAuthUser(
            id="99",
            username="userone",
            email="user@example.com",
        )


def test_github_provider(db, settings):
    settings.OAUTH_LOGIN_PROVIDERS = {
        "github": {
            "class": "provider_tests.test_github.DummyGitHubOAuthProvider",
            "kwargs": {
                "client_id": "test_id",
                "client_secret": "test_secret",
                "scope": "user",
            },
        }
    }

    client = Client()

    # Login required for this view
    response = client.get("/")
    assert response.status_code == 302
    assert response.url == "/login/?next=/"

    # User clicks the login link (form submit)
    response = client.post("/oauth/github/login/")
    assert response.status_code == 302
    assert (
        response.url
        == "https://github.com/login/oauth/authorize?client_id=test_id&redirect_uri=https%3A%2F%2Ftestserver%2Foauth%2Fgithub%2Fcallback%2F&response_type=code&scope=user&state=dummy_state"
    )

    # GitHub redirects to the callback url
    response = client.get("/oauth/github/callback/?code=test_code&state=dummy_state")
    assert response.status_code == 302
    assert response.url == "/"

    # Now logged in
    response = client.get("/")
    assert response.status_code == 200
    assert b"Hello userone!\n" in response.content

    # Check the user and connection that was created
    user = response.user
    assert user.username == "userone"
    assert user.email == "user@example.com"
    connections = user.oauth_connections.all()
    assert len(connections) == 1
    assert connections[0].provider_key == "github"
    assert connections[0].provider_user_id == "99"
    assert connections[0].access_token == "gho_key"
    assert connections[0].refresh_token == ""
    assert connections[0].access_token_expires_at is None
